@@ -1,11 +1,11 @@
 /*
  *	libHX/opt.c
- *	Copyright © Jan Engelhardt <jengelh [at] medozas de>, 2002 - 2011
+ *	Copyright Jan Engelhardt, 2002-2011
  *
  *	This file is part of libHX. libHX is free software; you can
- *	redistribute it and/or modify it under the terms of the GNU
- *	Lesser General Public License as published by the Free Software
- *	Foundation; either version 2.1 or 3 of the License.
+ *	redistribute it and/or modify it under the terms of the GNU Lesser
+ *	General Public License as published by the Free Software Foundation;
+ *	either version 2.1 or (at your option) any later version.
  */
 #include <errno.h>
 #include <stdbool.h>
@@ -61,8 +61,6 @@ enum {
 	HXOPT_LOPMASK2 = HXOPT_OR | HXOPT_AND | HXOPT_XOR,
 	HXOPT_LOPMASK  = HXOPT_LOPMASK2 | HXOPT_NOT,
 	HXOPT_TYPEMASK = 0x1F, /* 5 bits */
-
-	HXOPT_POSIX_MODE = 1 << 16,
 };
 
 /**
@@ -72,6 +70,7 @@ enum {
  * %HXOPT_E_LONG_MISSING:	long option requires an argument
  * %HXOPT_E_SHORT_UNKNOWN:	unknown short option
  * %HXOPT_E_SHORT_MISSING:	short option requires an argument
+ * %HXOPT_E_AMBIG_PREFIX:	used abbreviated long option but had multiple results
  */
 enum {
 	HXOPT_E_LONG_UNKNOWN = 1,
@@ -79,6 +78,7 @@ enum {
 	HXOPT_E_LONG_MISSING,
 	HXOPT_E_SHORT_UNKNOWN,
 	HXOPT_E_SHORT_MISSING,
+	HXOPT_E_AMBIG_PREFIX,
 };
 
 /**
@@ -129,6 +129,8 @@ struct HX_getopt_vars {
 	unsigned int flags;
 };
 
+struct HXoption HXopt_ambig_prefix;
+
 static bool posix_me_harder(void)
 {
 	const char *s;
@@ -200,6 +202,7 @@ static void do_assign(struct HXoptcb *cbi, const char *arg0)
 	NTYPE(HXTYPE_UINT64,   uint64_t, strtoull);
 	NTYPE(HXTYPE_INT64,    int64_t, strtoll);
 #endif
+	NTYPE(HXTYPE_SIZE_T,   size_t, strtoull);
 	case HXTYPE_FLOAT:
 		cbi->data_dbl = strtod(cbi->data, NULL);
 		if (opt->ptr != NULL)
@@ -233,8 +236,8 @@ static void do_assign(struct HXoptcb *cbi, const char *arg0)
 		opt->cb(cbi);
 }
 
-static inline const struct HXoption *lookup_short(const struct HXoption *table,
-    char opt)
+static __inline__ const struct HXoption *
+lookup_short(const struct HXoption *table, char opt)
 {
 	for (; table->type != HXTYPE_XSNTMARK; ++table)
 		if (table->sh == opt)
@@ -242,8 +245,29 @@ static inline const struct HXoption *lookup_short(const struct HXoption *table,
 	return NULL;
 }
 
-static inline const struct HXoption *lookup_long(const struct HXoption *table,
-    const char *key)
+static const struct HXoption *
+lookup_long_pfx(const struct HXoption *table, const char *key)
+{
+	const struct HXoption *cand = NULL;
+	size_t klen = strlen(key);
+
+	for (; table->type != HXTYPE_XSNTMARK; ++table) {
+		if (table->ln == NULL)
+			continue;
+		if (strncmp(table->ln, key, klen) != 0)
+			continue;
+		/* Prefix match */
+		if (table->ln[klen] == '\0')
+			return table; /* Exact match */
+		if (cand != NULL)
+			return &HXopt_ambig_prefix;
+		cand = table;
+	}
+	return cand;
+}
+
+static __inline__ const struct HXoption *
+lookup_long(const struct HXoption *table, const char *key)
 {
 	for (; table->type != HXTYPE_XSNTMARK; ++table)
 		if (table->ln != NULL && strcmp(table->ln, key) == 0)
@@ -251,7 +275,7 @@ static inline const struct HXoption *lookup_long(const struct HXoption *table,
 	return NULL;
 }
 
-static inline bool takes_void(unsigned int t)
+static __inline__ bool takes_void(unsigned int t)
 {
 	t &= HXOPT_TYPEMASK;
 	return t == HXTYPE_NONE || t == HXTYPE_VAL || t == HXTYPE_SVAL ||
@@ -280,22 +304,20 @@ static void opt_to_text(const struct HXoption *opt, char *buf, size_t len,
 		buf[i++] = opt->sh;
 		if (!takes_void(opt->type))
 			i += snprintf(buf + i, len - i, " %s", htyp);
+	} else if (opt->sh == '\0') {
+		if (takes_void(opt->type))
+			i += snprintf(buf + i, len - i,
+			     "--%s", opt->ln);
+		else
+			i += snprintf(buf + i, len - i,
+			     "--%s=%s", opt->ln, htyp);
 	} else {
-		if (opt->sh == '\0') {
-			if (takes_void(opt->type))
-				i += snprintf(buf + i, len - i,
-				     "--%s", opt->ln);
-			else
-				i += snprintf(buf + i, len - i,
-				     "--%s=%s", opt->ln, htyp);
-		} else {
-			if (takes_void(opt->type))
-				i += snprintf(buf + i, len - i, "-%c%s--%s",
-				     opt->sh, alt, opt->ln);
-			else
-				i += snprintf(buf + i, len - i, "-%c%s--%s%c%s",
-				     opt->sh, alt, opt->ln, equ, htyp);
-		}
+		if (takes_void(opt->type))
+			i += snprintf(buf + i, len - i, "-%c%s--%s",
+			     opt->sh, alt, opt->ln);
+		else
+			i += snprintf(buf + i, len - i, "-%c%s--%s%c%s",
+			     opt->sh, alt, opt->ln, equ, htyp);
 	}
 
 	if (flags & W_BRACKET)
@@ -326,7 +348,7 @@ static void print_indent(const char *msg, unsigned int ind, FILE *fp)
 }
 
 /**
- * HXparse_deshell_int - shell-style argument unescape
+ * HXparse_dequote_int - shell-style argument unescape
  * @o:		input/output string
  * @end:	terminating characters
  *
@@ -463,6 +485,10 @@ static int HX_getopt_error(int err, const char *key, unsigned int flags)
 			fprintf(stderr, "Option -%c requires an "
 			        "argument\n", *key);
 		return HXOPT_I_ERROR | HXOPT_ERR_MIS;
+	case HXOPT_E_AMBIG_PREFIX:
+		if (!(flags & HXOPT_QUIET))
+			fprintf(stderr, "Option %s is ambiguous\n", key);
+		return HXOPT_I_ERROR | HXOPT_ERR_AMBIG;
 	}
 	return HXOPT_I_ERROR;
 }
@@ -472,7 +498,9 @@ static int HX_getopt_twolong(const char *const *opt,
 {
 	const char *key = opt[0], *value = opt[1];
 
-	par->cbi.current = lookup_long(par->cbi.table, key + 2);
+	par->cbi.current = lookup_long_pfx(par->cbi.table, key + 2);
+	if (par->cbi.current == &HXopt_ambig_prefix)
+		return HX_getopt_error(HXOPT_E_AMBIG_PREFIX, key, par->flags);
 	if (par->cbi.current == NULL) {
 		if (par->flags & HXOPT_PTHRU) {
 			char *tmp = HX_strdup(key);
@@ -522,7 +550,12 @@ static int HX_getopt_long(const char *cur, struct HX_getopt_vars *par)
 
 	value = strchr(key, '=');
 	*value++ = '\0';
-	par->cbi.current = lookup_long(par->cbi.table, key + 2);
+	par->cbi.current = lookup_long_pfx(par->cbi.table, key + 2);
+	if (par->cbi.current == &HXopt_ambig_prefix) {
+		ret = HX_getopt_error(HXOPT_E_AMBIG_PREFIX, key, par->flags);
+		free(key);
+		return ret;
+	}
 	if (par->cbi.current == NULL) {
 		if (par->flags & HXOPT_PTHRU) {
 			/* Undo nuke of '=' and reuse alloc */
@@ -651,7 +684,7 @@ static int HX_getopt_normal(const char *cur, const struct HX_getopt_vars *par)
 	if (cur[0] == '-')
 		/* Short option(s) - one or more(!) */
 		return HXOPT_S_SHORT | HXOPT_I_ADVCHAR;
-	if (par->flags & HXOPT_POSIX_MODE)
+	if (par->flags & HXOPT_RQ_ORDER)
 		/* POSIX: first non-option implies option termination */
 		return HXOPT_S_TERMINATED;
 	cur = HX_strdup(cur);
@@ -690,7 +723,7 @@ EXPORT_SYMBOL int HX_getopt(const struct HXoption *table, int *argc,
 	}
 
 	if (posix_me_harder())
-		ps.flags |= HXOPT_POSIX_MODE;
+		ps.flags |= HXOPT_RQ_ORDER;
 
 	for (cur = *opt; cur != NULL; ) {
 		if (state == HXOPT_S_TWOLONG)
@@ -725,6 +758,7 @@ EXPORT_SYMBOL int HX_getopt(const struct HXoption *table, int *argc,
 
  out:
 	if (ret == HXOPT_ERR_SUCCESS) {
+		if (!(ps.flags & HXOPT_KEEP_ARGV)) {
 		const char **nvec = reinterpret_cast(const char **,
 		                    HXdeque_to_vec(ps.remaining, &argk));
 		if (nvec == NULL)
@@ -743,6 +777,7 @@ EXPORT_SYMBOL int HX_getopt(const struct HXoption *table, int *argc,
 		*argv = nvec;
 		if (argc != NULL)
 			*argc = argk;
+		}
 	} else if (ret < 0) {
 		if (!(ps.flags & HXOPT_QUIET))
 			fprintf(stderr, "%s: %s\n", __func__, strerror(errno));
@@ -806,7 +841,7 @@ EXPORT_SYMBOL void HX_getopt_usage(const struct HXoptcb *cbi, FILE *nfp)
 	/* Program name now expected in .data */
 	const char *arg0 = cbi->data;
 
-	if (arg0 == NULL)
+	if (arg0 == NULL || *arg0 == '\0')
 		arg0 = "($0)";
 
 	wd = sizeof("Usage:") + strlen(arg0);
@@ -942,7 +977,9 @@ EXPORT_SYMBOL struct HXmap *HX_shconfig_map(const char *file)
 		return NULL;
 
 	if ((fp = fopen(file, "r")) == NULL) {
-		free(map);
+		int saved_errno = errno;
+		HXmap_free(map);
+		errno = saved_errno;
 		return NULL;
 	}
 
@@ -976,6 +1013,8 @@ EXPORT_SYMBOL int HX_shconfig_pv(const char **path, const char *file,
 
 EXPORT_SYMBOL void HX_shconfig_free(const struct HXoption *table)
 {
+	if (table == NULL)
+		return;
 	for (; table->ln != NULL; ++table) {
 		char **ptr = table->ptr;
 		if (table->type == HXTYPE_STRING &&
