@@ -1,13 +1,15 @@
 /*
  *	C-string functions
- *	Copyright © Jan Engelhardt <jengelh [at] medozas de>, 1999 - 2010
+ *	Copyright Jan Engelhardt, 1999-2010
  *
  *	This file is part of libHX. libHX is free software; you can
- *	redistribute it and/or modify it under the terms of the GNU
- *	Lesser General Public License as published by the Free Software
- *	Foundation; either version 2.1 or 3 of the License.
+ *	redistribute it and/or modify it under the terms of the GNU Lesser
+ *	General Public License as published by the Free Software Foundation;
+ *	either version 2.1 or (at your option) any later version.
  */
 #include <errno.h>
+#include <limits.h>
+#include <math.h>
 #include <stdbool.h>
 #include <stddef.h>
 #include <stdint.h>
@@ -18,7 +20,29 @@
 #include <libHX/string.h>
 #include "internal.h"
 
-static inline unsigned int min_uint(unsigned int a, unsigned int b)
+/**
+ * %HXQUOTE_ACCEPT:	the listed characters are passed through,
+ * 			all others need to be quoted
+ * %HXQUOTE_REJECT:	the listed characters need to be quoted,
+ * 			all others pass through
+ */
+enum HX_quote_selector {
+	HXQUOTE_ACCEPT,
+	HXQUOTE_REJECT,
+};
+
+/**
+ * @selector:	whether this rule is accept- or reject-based
+ * @qchars:	characters that need (no) quoting
+ */
+struct HX_quote_rule {
+	char selector;
+	const char *chars;
+};
+
+static const char HX_hexenc[16] = "0123456789ABCDEF";
+
+static __inline__ unsigned int min_uint(unsigned int a, unsigned int b)
 {
 	return (a < b) ? a : b;
 }
@@ -26,9 +50,15 @@ static inline unsigned int min_uint(unsigned int a, unsigned int b)
 EXPORT_SYMBOL char *HX_basename(const char *s)
 {
 	const char *p;
+	/* ignore trailing slashes */
 	for (p = s + strlen(s) - 1; p >= s && *p == '/'; --p)
 		;
 	if (p < s)
+		/*
+		 * String contained only slashes - this must be the root
+		 * directory. Since we have the opportunity, rather than
+		 * returning "////", just give the cleaned-up "/".
+		 */
 		return const_cast1(char *, s + strlen(s) - 1);
 	if ((p = HX_strbchr(s, p, '/')) != NULL)
 		return const_cast1(char *, p + 1);
@@ -43,9 +73,11 @@ EXPORT_SYMBOL char *HX_basename_exact(const char *s)
 
 	if (*s == '\0')
 		return HX_strdup(".");
+	/* ignore trailing slashes */
 	for (end = s + strlen(s) - 1; end >= s && *end == '/'; --end)
 		;
 	if (end < s)
+		/* string consisted of only slashes */
 		return HX_strdup("/");
 
 	start = HX_strbchr(s, end, '/');
@@ -122,16 +154,33 @@ EXPORT_SYMBOL hxmc_t *HX_getl(hxmc_t **ptr, FILE *fp)
 	return *ptr;
 }
 
-EXPORT_SYMBOL void *HX_memmem(const void *space, size_t spacesize,
-    const void *point, size_t pointsize)
+EXPORT_SYMBOL void *HX_memmem(const void *vspace, size_t spacesize,
+    const void *vpoint, size_t pointsize)
 {
-	size_t i;
+	const char *space = vspace, *point = vpoint;
+	const char *head, *end;
+	size_t tailsize;
+	char *tail;
 
+	if (pointsize == 0)
+		return const_cast1(void *, vspace);
 	if (pointsize > spacesize)
 		return NULL;
-	for (i = 0; i <= spacesize - pointsize; ++i)
-		if (memcmp(space + i, point, pointsize) == 0)
-			return const_cast1(void *, space + i);
+
+	/* Do a BM-style trailer search and reduce calls to memcmp */
+	head = space + (pointsize - 1);
+	tail = memchr(head, point[pointsize-1], spacesize - (pointsize - 1));
+	if (tail == NULL || pointsize == 1)
+		return tail;
+	end = space + spacesize;
+	do {
+		head = tail - pointsize + 1;
+		if (memcmp(head, point, pointsize) == 0)
+			return const_cast(char *, head);
+		++tail;
+		tailsize = end - tail;
+		tail = memchr(tail, point[pointsize-1], tailsize);
+	} while (tail != NULL);
 	return NULL;
 }
 
@@ -185,7 +234,7 @@ EXPORT_SYMBOL char **HX_split(const char *str, const char *delim,
 	return ret;
 }
 
-EXPORT_SYMBOL char **HX_split4(char *s, const char *delim, int *fld, int max)
+EXPORT_SYMBOL char **HX_split_inplace(char *s, const char *delim, int *fld, int max)
 {
 	char **stk;
 	const char *p = s;
@@ -201,22 +250,22 @@ EXPORT_SYMBOL char **HX_split4(char *s, const char *delim, int *fld, int max)
 	if (stk == NULL)
 		return NULL;
 	stk[count] = NULL;
-	count = HX_split5(s, delim, count, stk);
+	count = HX_split_fixed(s, delim, count, stk);
 	if (fld != NULL)
 		*fld = count;
 	return stk;
 }
 
-EXPORT_SYMBOL int HX_split5(char *s, const char *delim, int max, char **stk)
+EXPORT_SYMBOL int HX_split_fixed(char *s, const char *delim, int max, char **stk)
 {
 	/*
-	 * HX_split5 - the "stack split" (we try to avoid using the heap):
+	 * HX_split_fixed - the "stack split" (we try to avoid using the heap):
 	 * Split @s (modifies it, so must be writable!) at @delim with at most
 	 * @max fields and putting the results into @stk[0..@max-1].
 	 *
 	 * Example on @max:
 	 *	char *stk[max];
-	 *	HX_split5(s, delim, max, stk);
+	 *	HX_split_fixed(s, delim, max, stk);
 	 */
 	int i = 0;
 	char *p;
@@ -242,6 +291,19 @@ EXPORT_SYMBOL char *HX_strbchr(const char *start, const char *now, char d)
 	return NULL;
 }
 
+/**
+ * This is the counterpart to strpbrk(). Returns a pointer to the first
+ * character not in @accept, or otherwise %NULL.
+ */
+EXPORT_SYMBOL char *HX_strchr2(const char *s, const char *accept)
+{
+	size_t seg = strspn(s, accept);
+
+	if (s[seg] == '\0')
+		return NULL;
+	return const_cast1(char *, s + seg);
+}
+
 EXPORT_SYMBOL char *HX_strclone(char **pa, const char *pb)
 {
 	if (*pa == pb)
@@ -260,6 +322,8 @@ EXPORT_SYMBOL char *HX_strclone(char **pa, const char *pb)
 
 EXPORT_SYMBOL char *HX_strdup(const char *src)
 {
+	if (src == NULL)
+		return NULL;
 	/* return HX_strndup(src, SIZE_MAX); */
 	return HX_memdup(src, strlen(src) + 1);
 }
@@ -274,6 +338,8 @@ EXPORT_SYMBOL char *HX_strlcat(char *dest, const char *src, size_t len)
 
 EXPORT_SYMBOL char *HX_strlcpy(char *dest, const char *src, size_t n)
 {
+	if (n == 0)
+		return dest;
 	strncpy(dest, src, n);
 	dest[n-1] = '\0';
 	return dest;
@@ -303,13 +369,20 @@ EXPORT_SYMBOL size_t HX_strltrim(char *expr)
 	size_t diff = 0;
 	travp = expr;
 
-	while (*travp != '\0' && HX_isspace(*travp))
+	while (HX_isspace(*travp))
 		++travp;
 	if ((diff = travp - expr) > 0)
-		memmove(expr, travp, diff);
+		memmove(expr, travp, strlen(travp) + 1);
 	return diff;
 }
 
+EXPORT_SYMBOL char *HX_stpltrim(const char *p)
+{
+	while (HX_isspace(*p))
+		++p;
+	return const_cast1(char *, p);
+}
+
 /* supports negative offsets like scripting languages */
 EXPORT_SYMBOL char *HX_strmid(const char *expr, long offset, long length)
 {
@@ -409,7 +482,7 @@ EXPORT_SYMBOL char *HX_strsep(char **sp, const char *d)
 		*end++ = '\0';
 		*sp = end;
 	}
-	
+
 	return begin;
 }
 
@@ -428,24 +501,51 @@ EXPORT_SYMBOL char *HX_strsep2(char **wp, const char *str)
 	return ret;
 }
 
-static const char *const HX_quote_chars[] = {
-	[HXQUOTE_SQUOTE] = "'\\",
-	[HXQUOTE_DQUOTE] = "\"\\",
-	[HXQUOTE_HTML]   = "\"&<>",
-	[HXQUOTE_LDAPFLT] = "\n*()\\",
-	[HXQUOTE_LDAPRDN] = "\n \"#+,;<=>\\",
+static const struct HX_quote_rule HX_quote_rules[] = {
+	[HXQUOTE_SQUOTE]  = {HXQUOTE_REJECT, "'\\"},
+	[HXQUOTE_DQUOTE]  = {HXQUOTE_REJECT, "\"\\"},
+	[HXQUOTE_HTML]    = {HXQUOTE_REJECT, "\"&<>"},
+	[HXQUOTE_LDAPFLT] = {HXQUOTE_REJECT, "\n*()\\"},
+	[HXQUOTE_LDAPRDN] = {HXQUOTE_REJECT, "\n \"#+,;<=>\\"},
+	[HXQUOTE_URIENC]  = {HXQUOTE_ACCEPT, "-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz~"},
+	[HXQUOTE_SQLSQUOTE] = {HXQUOTE_REJECT, "'"},
+	[HXQUOTE_SQLBQUOTE] = {HXQUOTE_REJECT, "`"},
 };
 
 /**
- * HX_qsize_backslash - calculate size of new buffer
+ * HX_qsize_bsa - calculate length of statically expanded string (for accepts)
  * @s:		input string
  * @qchars:	characters that need quoting
  * @cost:	quoting cost per quoted character
  *
- * The cost depends on the quote format (\a vs \07 vs \x07).
+ * The cost depends on the quote format. Typical values:
+ * 	1	when "&" becomes "\&" (programming language-like)
+ * 	2	when "&" becomes "\26" (LDAPRDN/HTTPURI-like hex encoding)
+ * 	3	when "&" becomes "\x26" (hex encoding for programming)
  */
-static size_t HX_qsize_backslash(const char *s, const char *qchars,
-    unsigned int cost)
+static size_t
+HX_qsize_bsa(const char *s, const char *qchars, unsigned int cost)
+{
+	const char *p = s;
+	size_t n = strlen(s);
+
+	while ((p = HX_strchr2(p, qchars)) != NULL) {
+		n += cost;
+		++p;
+	}
+	return n;
+}
+
+/**
+ * HX_qsize_bsr - calculate length of statically expanded string (for rejects)
+ * @s:		input string
+ * @qchars:	characters that need quoting
+ * @cost:	quoting cost per quoted character
+ *
+ * Same as for HX_qsize_bsa, but for HXQUOTE_REJECT-type rules.
+ */
+static size_t
+HX_qsize_bsr(const char *s, const char *qchars, unsigned int cost)
 {
 	const char *p = s;
 	size_t n = strlen(s);
@@ -479,6 +579,30 @@ static char *HX_quote_backslash(char *dest, const char *src, const char *qc)
 	return ret;
 }
 
+static char *
+HX_quote_sqlbackslash(char *dest, const char *src, const char *trm)
+{
+	char *ret = dest;
+	size_t len;
+
+	while (*src != '\0') {
+		len = strcspn(src, trm);
+		if (len > 0) {
+			memcpy(dest, src, len);
+			dest += len;
+			src  += len;
+			if (*src == '\0')
+				break;
+		}
+		*dest++ = *trm;
+		*dest++ = *trm;
+		++src;
+	}
+
+	*dest = '\0';
+	return ret;
+}
+
 /**
  * Encode @src into BASE-64 according to RFC 4648 and write result to @dest,
  * which must be of appropriate size, plus one for a trailing NUL.
@@ -523,7 +647,7 @@ static size_t HX_qsize_html(const char *s)
 	const char *p = s;
 	size_t n = strlen(s);
 
-	while ((p = strpbrk(p, HX_quote_chars[HXQUOTE_HTML])) != NULL) {
+	while ((p = strpbrk(p, HX_quote_rules[HXQUOTE_HTML].chars)) != NULL) {
 		switch (*p) {
 		/* minus 2: \0 and the original char */
 		case '"':
@@ -552,7 +676,7 @@ static char *HX_quote_html(char *dest, const char *src)
 	char *ret = dest;
 
 	while (*src != '\0') {
-		size_t len = strcspn(src, HX_quote_chars[HXQUOTE_HTML]);
+		size_t len = strcspn(src, HX_quote_rules[HXQUOTE_HTML].chars);
 		if (len > 0) {
 			memcpy(dest, src, len);
 			dest += len;
@@ -574,7 +698,6 @@ static char *HX_quote_html(char *dest, const char *src)
 
 static char *HX_quote_ldap(char *dest, const char *src, const char *qc)
 {
-	static const char hex[] = "0123456789ABCDEF";
 	char *ret = dest;
 	size_t len;
 
@@ -588,14 +711,36 @@ static char *HX_quote_ldap(char *dest, const char *src, const char *qc)
 				break;
 		}
 		*dest++ = '\\';
-		*dest++ = hex[(*src >> 4) & 0x0F];
-		*dest++ = hex[*src++ & 0x0F];
+		*dest++ = HX_hexenc[(*src >> 4) & 0x0F];
+		*dest++ = HX_hexenc[*src++ & 0x0F];
 	}
 
 	*dest = '\0';
 	return ret;
 }
 
+static char *HX_quote_urlenc(char *dest, const char *src)
+{
+	char *ret = dest;
+	size_t len;
+
+	while (*src != '\0') {
+		len = strspn(src, HX_quote_rules[HXQUOTE_URIENC].chars);
+		if (len > 0) {
+			memcpy(dest, src, len);
+			dest += len;
+			src  += len;
+			if (*src == '\0')
+				break;
+		}
+		*dest++ = '%';
+		*dest++ = HX_hexenc[(*src >> 4) & 0x0F];
+		*dest++ = HX_hexenc[*src++ & 0x0F];
+	}
+
+	*dest = '\0';
+	return ret;
+}
 
 /**
  * HX_quoted_size -
@@ -609,14 +754,18 @@ static size_t HX_quoted_size(const char *s, unsigned int type)
 	switch (type) {
 	case HXQUOTE_SQUOTE:
 	case HXQUOTE_DQUOTE:
-		return HX_qsize_backslash(s, HX_quote_chars[type], 1);
+	case HXQUOTE_SQLSQUOTE:
+	case HXQUOTE_SQLBQUOTE:
+		return HX_qsize_bsr(s, HX_quote_rules[type].chars, 1);
 	case HXQUOTE_HTML:
 		return HX_qsize_html(s);
 	case HXQUOTE_LDAPFLT:
 	case HXQUOTE_LDAPRDN:
-		return HX_qsize_backslash(s, HX_quote_chars[type], 2);
+		return HX_qsize_bsr(s, HX_quote_rules[type].chars, 2);
 	case HXQUOTE_BASE64:
 		return (strlen(s) + 2) / 3 * 4;
+	case HXQUOTE_URIENC:
+		return HX_qsize_bsa(s, HX_quote_rules[type].chars, 2);
 	default:
 		return strlen(s);
 	}
@@ -625,6 +774,7 @@ static size_t HX_quoted_size(const char *s, unsigned int type)
 EXPORT_SYMBOL char *HX_strquote(const char *src, unsigned int type,
     char **free_me)
 {
+	const struct HX_quote_rule *rule;
 	bool do_quote;
 	char *tmp;
 
@@ -633,9 +783,15 @@ EXPORT_SYMBOL char *HX_strquote(const char *src, unsigned int type,
 		return NULL;
 	}
 	/* If quote_chars is NULL, it is clear all chars are to be encoded. */
-	do_quote = type >= ARRAY_SIZE(HX_quote_chars) ||
-	           HX_quote_chars[type] == NULL ||
-	           strpbrk(src, HX_quote_chars[type]) != NULL;
+	rule = &HX_quote_rules[type];
+	if (type >= ARRAY_SIZE(HX_quote_rules) || rule->chars == NULL)
+		do_quote = true;
+	else if (rule->selector == HXQUOTE_REJECT)
+		do_quote = strpbrk(src, rule->chars) != NULL;
+	else if (rule->selector == HXQUOTE_ACCEPT)
+		do_quote = HX_strchr2(src, rule->chars) != NULL;
+	else
+		do_quote = false;
 	/*
 	 * free_me == NULL implies that we always allocate, even if
 	 * there is nothing to quote.
@@ -658,14 +814,19 @@ EXPORT_SYMBOL char *HX_strquote(const char *src, unsigned int type,
 	switch (type) {
 	case HXQUOTE_SQUOTE:
 	case HXQUOTE_DQUOTE:
-		return HX_quote_backslash(*free_me, src, HX_quote_chars[type]);
+		return HX_quote_backslash(*free_me, src, rule->chars);
 	case HXQUOTE_HTML:
 		return HX_quote_html(*free_me, src);
 	case HXQUOTE_LDAPFLT:
 	case HXQUOTE_LDAPRDN:
-		return HX_quote_ldap(*free_me, src, HX_quote_chars[type]);
+		return HX_quote_ldap(*free_me, src, rule->chars);
 	case HXQUOTE_BASE64:
 		return HX_quote_base64(*free_me, src);
+	case HXQUOTE_URIENC:
+		return HX_quote_urlenc(*free_me, src);
+	case HXQUOTE_SQLSQUOTE:
+	case HXQUOTE_SQLBQUOTE:
+		return HX_quote_sqlbackslash(*free_me, src, rule->chars);
 	}
 	return NULL;
 }
@@ -677,3 +838,169 @@ EXPORT_SYMBOL char *HX_strupper(char *orig)
 		*expr = HX_toupper(*expr);
 	return orig;
 }
+
+EXPORT_SYMBOL char *HX_unit_size(char *buf, size_t bufsize,
+    unsigned long long size, unsigned int divisor, unsigned int cutoff)
+{
+	static const char unit_names[] = "\0kMGTPEZYRQ";
+	unsigned int unit_idx = 0;
+	if (divisor == 0)
+		divisor = 1000;
+	if (cutoff == 0) {
+		cutoff = 10000;
+		if (cutoff < divisor)
+			cutoff = divisor;
+	}
+	while (unit_idx < ARRAY_SIZE(unit_names) - 1 && size >= cutoff) {
+		++unit_idx;
+		size /= divisor;
+	}
+	snprintf(buf, bufsize, "%llu%.1s", size, &unit_names[unit_idx]);
+	return buf;
+}
+
+static inline unsigned long long p_90(unsigned long long x)
+{
+	/* Perform x*9/10, but without the risk of overflow. */
+	return x - x / 10 - !!(x % 10);
+}
+
+EXPORT_SYMBOL char *HX_unit_size_cu(char *buf, size_t bufsize,
+    unsigned long long orig_size, unsigned int divisor)
+{
+	/* No floating point. Take that, coreutils! */
+	static const char unit_names[] = "\0kMGTPEZYRQ";
+	unsigned int unit_idx = 0, last_rem = 0;
+	unsigned long long size = orig_size, gpow = 1, grand_rem;
+	if (divisor == 0)
+		divisor = 1000;
+
+	while (unit_idx < ARRAY_SIZE(unit_names) - 1 && size >= divisor) {
+		++unit_idx;
+		last_rem = size % divisor;
+		size /= divisor;
+		gpow *= divisor;
+	}
+	if (unit_idx == 0) {
+		snprintf(buf, bufsize, "%llu%.1s", size, &unit_names[unit_idx]);
+		return buf;
+	}
+	grand_rem = orig_size - size * gpow;
+	if (grand_rem != 0) {
+		if (grand_rem > p_90(gpow)) {
+			++size;
+			last_rem = 0;
+		} else {
+			last_rem *= 10;
+			last_rem /= divisor;
+			++last_rem;
+			if (last_rem == 10 || (size >= 10 && last_rem > 0)) {
+				++size;
+				last_rem = 0;
+			}
+		}
+		if (unit_idx < ARRAY_SIZE(unit_names) - 1 && size == divisor) {
+			/* ++size from above may brought size to @divisor again */
+			++unit_idx;
+			size /= divisor;
+		}
+	}
+	if (size >= 10 && last_rem == 0)
+		snprintf(buf, bufsize, "%llu%.1s", size, &unit_names[unit_idx]);
+	else
+		snprintf(buf, bufsize, "%llu.%01u%.1s", size, last_rem, &unit_names[unit_idx]);
+	return buf;
+}
+
+static unsigned int suffix_power(char u)
+{
+	switch (HX_toupper(u)) {
+	case 'K': return 1;
+	case 'M': return 2;
+	case 'G': return 3;
+	case 'T': return 4;
+	case 'P': return 5;
+	case 'E': return 6;
+	case 'Z': return 7;
+	case 'Y': return 8;
+	case 'R': return 9;
+	case 'Q': return 10;
+	default: return 0;
+	}
+}
+
+EXPORT_SYMBOL double HX_strtod_unit(const char *s, char **out_end, unsigned int exponent)
+{
+	char *end;
+	double q;
+
+	while (HX_isspace(*s))
+		++s;
+	if (*s == '-') {
+		if (out_end != nullptr)
+			*out_end = const_cast(char *, s);
+		errno = ERANGE;
+		return ULLONG_MAX;
+	}
+	q = strtod(s, &end);
+	if (exponent == 0)
+		exponent = 1000;
+	if (end == s) {
+		if (out_end != nullptr)
+			*out_end = end;
+		return q;
+	}
+	while (HX_isspace(*end))
+		++end;
+	unsigned int pwr = suffix_power(*end);
+	if (pwr == 0) {
+		if (out_end != nullptr)
+			*out_end = const_cast(char *, end);
+		return q;
+	}
+	if (out_end != nullptr)
+		*out_end = const_cast(char *, end + 1);
+	return q * pow(exponent, pwr);
+}
+
+EXPORT_SYMBOL unsigned long long HX_strtoull_unit(const char *s,
+    char **out_end, unsigned int exponent)
+{
+	char *end;
+	unsigned long long ipart;
+	unsigned int pwr = 0;
+
+	while (HX_isspace(*s))
+		++s;
+	if (*s == '-') {
+		if (out_end != nullptr)
+			*out_end = const_cast(char *, s);
+		errno = ERANGE;
+		return ULLONG_MAX;
+	}
+	ipart = strtoull(s, &end, 10);
+	if (*end == '.') {
+		double q = HX_strtod_unit(s, out_end, exponent);
+		return q < ULLONG_MAX ? q : ULLONG_MAX;
+	}
+	if (exponent == 0)
+		exponent = 1000;
+	while (HX_isspace(*end))
+		++end;
+	pwr = suffix_power(*end);
+	if (pwr == 0) {
+		if (out_end != nullptr)
+			*out_end = end;
+		return ipart;
+	}
+	if (out_end != nullptr)
+		*out_end = const_cast(char *, end + 1);
+	while (pwr-- > 0) {
+		if (ipart >= ULLONG_MAX / exponent) {
+			errno = ERANGE;
+			return ULLONG_MAX;
+		}
+		ipart *= exponent;
+	}
+	return ipart;
+}
